@@ -1,7 +1,3 @@
-# A general class for utility functions.  Please use this for
-# functions that stand alone, without requiring a $self object,
-# Portability functions especially.
-
 # <@LICENSE>
 # Copyright 2004 Apache Software Foundation
 # 
@@ -18,11 +14,33 @@
 # limitations under the License.
 # </@LICENSE>
 
+=head1 NAME
+
+Mail::SpamAssassin::Util - utility functions
+
+=head1 DESCRIPTION
+
+A general class for utility functions.  Please use this for functions that
+stand alone, without requiring a $self object, Portability functions
+especially.
+
+NOTE: The functions in this module are to be considered private.  Their API may
+change at any point, and it's expected that they'll only be used by other
+Mail::SpamAssassin modules. (TODO: we should probably revisit this if
+it's useful for plugin development.)
+
+=over 4
+
+=cut
+
 package Mail::SpamAssassin::Util;
 
 use strict;
+use warnings;
 use bytes;
 
+use Mail::SpamAssassin::Logger;
+
 use vars qw (
   @ISA @EXPORT
   $AM_TAINTED
@@ -42,6 +60,8 @@ use Time::Local;
 use Sys::Hostname (); # don't import hostname() into this namespace!
 use Fcntl;
 use POSIX (); # don't import anything unless we ask explicitly!
+use Text::Wrap ();
+use Errno qw(EEXIST);
 
 ###########################################################################
 
@@ -60,17 +80,17 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
 
     clean_path_in_taint_mode();
     if ( !$displayed_path++ ) {
-      dbg("Current PATH is: ".join($Config{'path_sep'},File::Spec->path()));
+      dbg("util: current PATH is: ".join($Config{'path_sep'},File::Spec->path()));
     }
     foreach my $path (File::Spec->path()) {
       my $fname = File::Spec->catfile ($path, $filename);
       if ( -f $fname ) {
         if (-x $fname) {
-          dbg ("executable for $filename was found at $fname");
+          dbg("util: executable for $filename was found at $fname");
           return $fname;
         }
         else {
-          dbg("$filename was found at $fname, but isn't executable");
+          dbg("util: $filename was found at $fname, but isn't executable");
         }
       }
     }
@@ -86,10 +106,15 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
   my $cleaned_taint_path = 0;
 
   sub clean_path_in_taint_mode {
-    return if ( $cleaned_taint_path++ );
+    return if ($cleaned_taint_path++);
     return unless am_running_in_taint_mode();
 
-    dbg("Running in taint mode, removing unsafe env vars, and resetting PATH");
+    dbg("util: taint mode: deleting unsafe environment variables, resetting PATH");
+
+    if (RUNNING_ON_WINDOWS) {
+      dbg("util: running on Win32, skipping PATH cleaning");
+      return;
+    }
 
     delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
 
@@ -103,31 +128,31 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
       $dir = File::Spec->canonpath($1);
 
       if (!File::Spec->file_name_is_absolute($dir)) {
-	dbg("PATH included '$dir', which is not absolute, dropping.");
+	dbg("util: PATH included '$dir', which is not absolute, dropping");
 	next;
       }
       elsif (!(@stat=stat($dir))) {
-	dbg("PATH included '$dir', which doesn't exist, dropping.");
+	dbg("util: PATH included '$dir', which doesn't exist, dropping");
 	next;
       }
       elsif (!-d _) {
-	dbg("PATH included '$dir', which isn't a directory, dropping.");
+	dbg("util: PATH included '$dir', which isn't a directory, dropping");
 	next;
       }
       elsif (($stat[2]&2) != 0) {
         # World-Writable directories are considered insecure.
         # We could be more paranoid and check all of the parent directories as well,
         # but it's good for now.
-	dbg("PATH included '$dir', which is world writable, dropping.");
+	dbg("util: PATH included '$dir', which is world writable, dropping");
 	next;
       }
 
-      dbg("PATH included '$dir', keeping.");
+      dbg("util: PATH included '$dir', keeping");
       push(@path, $dir);
     }
 
     $ENV{'PATH'} = join($Config{'path_sep'}, @path);
-    dbg("Final PATH set to: ".$ENV{'PATH'});
+    dbg("util: final PATH set to: ".$ENV{'PATH'});
   }
 }
 
@@ -156,7 +181,7 @@ sub am_running_in_taint_mode {
     # seriously mind-bending perl
     $AM_TAINTED = not eval { eval "1 || $blank" || 1 };
   }
-  dbg ("running in taint mode? ". ($AM_TAINTED ? "yes" : "no"));
+  dbg("util: running in taint mode? ". ($AM_TAINTED ? "yes" : "no"));
   return $AM_TAINTED;
 }
 
@@ -190,7 +215,7 @@ sub untaint_file_path {
   if ($path =~ $re) {
     return $1;
   } else {
-    warn "security: cannot untaint path: \"$path\"\n";
+    warn "util: cannot untaint path: \"$path\"\n";
     return $path;
   }
 }
@@ -201,15 +226,16 @@ sub untaint_hostname {
   return unless defined($host);
   return '' if ($host eq '');
 
-  # from RFC 1035, but allowing domains starting with numbers
-  my $label = q/[A-Za-z\d](?:[A-Za-z\d-]{0,61}[A-Za-z\d])?/;
-  my $domain = qq<$label(?:\.$label)*>;
-
-  if (length($host) <= 255 && $host =~ /^($domain)$/) {
+  # from RFC 1035, but allowing domains starting with numbers:
+  #   $label = q/[A-Za-z\d](?:[A-Za-z\d-]{0,61}[A-Za-z\d])?/;
+  #   $domain = qq<$label(?:\.$label)*>;
+  #   length($host) <= 255 && $host =~ /^($domain)$/
+  # expanded (no variables in the re) because of a tainting bug in Perl 5.8.0
+  if (length($host) <= 255 && $host =~ /^([a-z\d](?:[a-z\d-]{0,61}[a-z\d])?(?:\.[a-z\d](?:[a-z\d-]{0,61}[a-z\d])?)*)$/i) {
     return $1;
   }
   else {
-    warn "security: cannot untaint hostname: \"$host\"\n";
+    warn "util: cannot untaint hostname: \"$host\"\n";
     return $host;
   }
 }
@@ -251,7 +277,7 @@ sub untaint_var {
     ${$_} = untaint_var(${$_});
   }
   else {
-    warn "Can't untaint a " . ref($_) . "!\n";
+    warn "util: can't untaint a " . ref($_) . "!\n";
   }
   return $_;
 }
@@ -349,7 +375,7 @@ sub parse_rfc822_date {
   } elsif (s/ (\d+) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) (\d{2,3}) / /i) {
     $dd = $1; $mon = lc($2); $yyyy = $3;
   } else {
-    dbg ("time cannot be parsed: $date");
+    dbg("util: time cannot be parsed: $date");
     return undef;
   }
 
@@ -390,12 +416,12 @@ sub parse_rfc822_date {
   # Time::Local (v1.10 at least) throws warnings when the dates cause
   # a 32-bit overflow.  So force a min/max for year.
   if ($yyyy > 2037) {
-    dbg("util: forcing year to 2037: $date");
-    $yyyy=2037;
+    dbg("util: date after supported range, forcing year to 2037: $date");
+    $yyyy = 2037;
   }
   elsif ($yyyy < 1970) {
-    dbg("util: forcing year to 1970: $date");
-    $yyyy=1970;
+    dbg("util: date before supported range, forcing year to 1970: $date");
+    $yyyy = 1971;
   }
 
   my $time;
@@ -404,7 +430,7 @@ sub parse_rfc822_date {
   };
 
   if ($@) {
-    dbg ("time cannot be parsed: $date, $yyyy-$mmm-$dd $hh:$mm:$ss");
+    dbg("util: time cannot be parsed: $date, $yyyy-$mmm-$dd $hh:$mm:$ss");
     return undef;
   }
 
@@ -435,11 +461,55 @@ sub time_to_rfc822_date {
 
 ###########################################################################
 
+# This is a wrapper for the Text::Wrap::wrap routine which makes its usage
+# a bit safer.  It accepts values for almost all options which can be set
+# in Text::Wrap.   All parameters are optional (leaving away the first one 
+# probably doesn't make too much sense though), either a missing or a false
+# value will fall back to the default.  Note that the parameter order and
+# default values aren't always the isame as in Text::Wrap itself.
+# 
+# The parameters are:
+#  1st:  The string to wrap.  Only one string is allowed (unlike the original
+#        wrap() routine).  (default: "")
+#  2nd:  The prefix to be put in front of all lines except the first one. 
+#        (default: "")
+#  3rd:  The prefix for the first line.  (default:  "")
+#  4th:  The number of columns available (no line will be longer than this
+#        parameter minus one).  See $Text::Wrap::columns.  (default:  77)
+#  5th:  Enable or disable overflow mode.  A false value is 'overflow', a
+#        true one 'wrap'; see $Text::Wrap::huge.  (default: 0)
+#  6th:  The sequence/expression to wrap at.  See $Text::Wrap::break
+#        (default: '\s');
+#  7th:  The string to join the lines again.  See $Text::Wrap::separator.
+#        (default: "\n")
+#  8th:  All tabs (except any in the prefix strings) are first replaced
+#        with 8 spaces.  This parameter controls if any 8-space sequence 
+#        is replaced with tabs again later.  See $Text::Wrap::unexpand but
+#        note that we use a different default value.  (default: 0)
+
+sub wrap {
+  local($Text::Wrap::columns)   = $_[3] || 77;
+  local($Text::Wrap::huge)      = $_[4] ? 'overflow' : 'wrap';
+  local($Text::Wrap::break)     = $_[5] || '\s';
+  local($Text::Wrap::separator) = $_[6] || "\n";
+  local($Text::Wrap::unexpand)  = $_[7] || 0;
+  # There's a die() in there which "shouldn't happen", but better be
+  # paranoid.  We'll return the unwrapped string if anything went wrong.
+  my $text = $_[0] || "";
+  eval {
+    $text = Text::Wrap::wrap($_[2] || "", $_[1] || "", $text);
+  };
+  return $text;
+}
+
+###########################################################################
+
 # Some base64 decoders will remove intermediate "=" characters, others
 # will stop decoding on the first "=" character, this one translates "="
 # characters to null.
 sub base64_decode {
   local $_ = shift;
+  my $decoded_length = shift;
 
   s/\s+//g;
   if (HAS_MIME_BASE64 && (length($_) % 4 == 0) &&
@@ -447,11 +517,21 @@ sub base64_decode {
   {
     # only use MIME::Base64 when the XS and Perl are both correct and quiet
     s/(=+)(?!=*$)/'A' x length($1)/ge;
+
+    # If only a certain number of bytes are requested, truncate the encoded
+    # version down to the appropriate size and return the requested bytes
+    if (defined $decoded_length) {
+      $_ = substr $_, 0, 4 * (int($decoded_length/3) + 1);
+      my $decoded = MIME::Base64::decode_base64($_);
+      return substr $decoded, 0, $decoded_length;
+    }
+
+    # otherwise, just decode the whole thing and return it
     return MIME::Base64::decode_base64($_);
   }
-  tr|A-Za-z0-9+/=||cd;			# remove non-base64 characters
+  tr{A-Za-z0-9+/=}{}cd;			# remove non-base64 characters
   s/=+$//;				# remove terminating padding
-  tr|A-Za-z0-9+/=| -_`|;		# translate to uuencode
+  tr{A-Za-z0-9+/=}{ -_`};		# translate to uuencode
   s/.$// if (length($_) % 4 == 1);	# unpack cannot cope with extra byte
 
   my $length;
@@ -459,6 +539,13 @@ sub base64_decode {
   while ($_) {
     $length = (length >= 84) ? 84 : length;
     $out .= unpack("u", chr(32 + $length * 3/4) . substr($_, 0, $length, ''));
+    last if (defined $decoded_length && length $out >= $decoded_length);
+  }
+
+  # If only a certain number of bytes are requested, truncate the encoded
+  # version down to the appropriate size and return the requested bytes
+  if (defined $decoded_length) {
+    return substr $out, 0, $decoded_length;
   }
 
   return $out;
@@ -481,7 +568,7 @@ sub base64_encode {
 
   $_ = pack("u57", $_);
   s/^.//mg;
-  tr| -_`|A-Za-z0-9+/A|;
+  tr| -_`|A-Za-z0-9+/A|; # -> #`# <- kluge against vim syntax issues
   s/(A+)$/'=' x length $1/e;
   return $_;
 }
@@ -496,18 +583,18 @@ sub portable_getpwuid {
   if (!RUNNING_ON_WINDOWS) {
     eval ' sub _getpwuid_wrapper { getpwuid($_[0]); } ';
   } else {
-    dbg ("defining getpwuid() wrapper using 'unknown' as username");
-    eval ' sub _getpwuid_wrapper { fake_getpwuid($_[0]); } ';
+    dbg("util: defining getpwuid() wrapper using 'unknown' as username");
+    eval ' sub _getpwuid_wrapper { _fake_getpwuid($_[0]); } ';
   }
 
   if ($@) {
-    warn "Failed to define getpwuid() wrapper: $@\n";
+    warn "util: failed to define getpwuid() wrapper: $@\n";
   } else {
     return Mail::SpamAssassin::Util::_getpwuid_wrapper(@_);
   }
 }
 
-sub fake_getpwuid {
+sub _fake_getpwuid {
   return (
     'unknown',		# name,
     'x',		# passwd,
@@ -554,7 +641,7 @@ sub extract_ipv4_addr_from_string {
   my($hostname, $fq_hostname);
 
 # get the current host's unqalified domain name (better: return whatever
-# Sys::Hostname thinks out hostname is, might also be a full qualified one)
+# Sys::Hostname thinks our hostname is, might also be a full qualified one)
   sub hostname {
     return $hostname if defined($hostname);
 
@@ -669,9 +756,11 @@ sub parse_content_type {
   # Note: the header content may not be whitespace unfolded, so make sure the
   # REs do /s when appropriate.
   #
-  $ct =~ s/;.*$//s;                     # strip everything after first semi-colon
+  $ct =~ s/^\s+//;			# strip leading whitespace
+  $ct =~ s/;.*$//s;			# strip everything after first ';'
   $ct =~ s@^([^/]+(?:/[^/]*)?).*$@$1@s;	# only something/something ...
-  $ct =~ tr/\000-\040\177-\377\042\050\051\054\056\072-\077\100\133-\135//d;    # strip inappropriate chars
+  # strip inappropriate chars
+  $ct =~ tr/\000-\040\177-\377\042\050\051\054\056\072-\077\100\133-\135//d;
   $ct = lc $ct;
 
   # bug 4298: If at this point we don't have a content-type, assume text/plain
@@ -706,8 +795,8 @@ sub url_encode {
     }
     # other stuff
     else {
-      # 0x00-0x20, 0x7f-0xff, <, >, and " ... "
-      s/([\000-\040\177-\377\074\076\042])
+      # 0x00-0x20, 0x7f-0xff, ", %, <, >
+      s/([\000-\040\177-\377\042\045\074\076])
 	  /push(@encoded, $1) && sprintf "%%%02x", unpack("C",$1)/egx;
     }
   }
@@ -723,8 +812,8 @@ sub url_encode {
 
 =item $module = first_available_module (@module_list)
 
-Return the first module that can be successfully loaded with C<require>
-from the list.  Returns C<undef> if none are available.
+Return the name of the first module that can be successfully loaded with
+C<require> from the list.  Returns C<undef> if none are available.
 
 This is used instead of C<AnyDBM_File> as follows:
 
@@ -749,42 +838,64 @@ sub first_available_module {
 
 ###########################################################################
 
-# thanks to http://www2.picante.com:81/~gtaylor/autobuse/ for this
-# code.
+=item my ($filehandle, $filepath) = secure_tmpfile();
+
+Generates a filename for a temporary file, opens it exclusively and
+securely, and returns a filehandle to the open file (opened O_RDWR).
+
+If it cannot open a file after 20 tries, it returns C<undef>.
+
+=cut
+
+# thanks to http://www2.picante.com:81/~gtaylor/autobuse/ for this code
 sub secure_tmpfile {
-  my $tmpdir = Mail::SpamAssassin::Util::untaint_file_path(
-                 File::Spec->tmpdir()
-               );
+  my $tmpdir = Mail::SpamAssassin::Util::untaint_file_path(File::Spec->tmpdir());
+
   if (!$tmpdir) {
-    die "Cannot find a temporary directory! set TMP or TMPDIR in env";
+    # Note: we would prefer to keep this fatal, as not being able to
+    # find a writable tmpdir is a big deal for the calling code too.
+    # That would be quite a psychotic case, also.
+    warn "util: cannot find a temporary directory, set TMP or TMPDIR in environment";
+    return;
   }
 
-  my ($reportfile,$tmpfile);
+  my ($reportfile, $tmpfile);
   my $umask = umask 077;
-  do {
-    # we do not rely on the obscurity of this name for security...
+
+  for (my $retries = 20; $retries > 0; $retries--) {
+    # we do not rely on the obscurity of this name for security,
     # we use a average-quality PRG since this is all we need
-    my $suffix = join ('',
-                       (0..9, 'A'..'Z','a'..'z')[rand 62,
-                                                 rand 62,
-                                                 rand 62,
-                                                 rand 62,
-                                                 rand 62,
-                                                 rand 62]);
-    $reportfile = File::Spec->catfile(
-                    $tmpdir,
-                    join ('.',
-                      "spamassassin",
-                      $$,
-                      $suffix,
-                      "tmp",
-                    )
-                  );
-    # ...rather, we require O_EXCL|O_CREAT to guarantee us proper
-    # ownership of our file; read the open(2) man page.
-  } while (! sysopen ($tmpfile, $reportfile, O_RDWR|O_CREAT|O_EXCL, 0600));
+    my $suffix = join('', (0..9,'A'..'Z','a'..'z')[rand 62, rand 62, rand 62,
+						   rand 62, rand 62, rand 62]);
+    $reportfile = File::Spec->catfile($tmpdir,".spamassassin${$}${suffix}tmp");
+
+    # instead, we require O_EXCL|O_CREAT to guarantee us proper
+    # ownership of our file, read the open(2) man page
+    if (sysopen($tmpfile, $reportfile, O_RDWR|O_CREAT|O_EXCL, 0600)) {
+      last;
+    }
+
+    if ($!{EEXIST}) {
+      # it is acceptable if $tmpfile already exists, try another
+      next;
+    }
+    
+    # error, maybe "out of quota" or "too many open files" (bug 4017)
+    warn "util: secure_tmpfile failed to create file '$tmpfile': $!\n";
+
+    # ensure the file handle is not semi-open in some way
+    if ($tmpfile) {
+      close $tmpfile;
+    }
+  }
+
   umask $umask;
 
+  if (!$tmpfile) {
+    warn "util: secure_tmpfile failed to create file, giving up";
+    return;	# undef
+  }
+
   return ($reportfile, $tmpfile);
 }
 
@@ -793,19 +904,23 @@ sub secure_tmpfile {
 sub uri_to_domain {
   my ($uri) = @_;
 
-  #return if ($uri =~ /^mailto:/i);	# not mailto's, please (TODO?)
-
   # Javascript is not going to help us, so return.
   return if ($uri =~ /^javascript:/i);
 
   $uri =~ s,#.*$,,gs;			# drop fragment
   $uri =~ s#^[a-z]+:/{0,2}##gsi;	# drop the protocol
   $uri =~ s,^[^/]*\@,,gs;		# username/passwd
-  $uri =~ s,[/\?].*$,,gs;		# path/cgi params
+
+  # strip path and CGI params.  note: bug 4213 shows that "&" should
+  # *not* be likewise stripped here -- it's permitted in hostnames by
+  # some common MUAs!
+  $uri =~ s,[/\?].*$,,gs;              
+
   $uri =~ s,:\d*$,,gs;			# port, bug 4191: sometimes the # is missing
 
-  return if $uri =~ /\%/;         # skip undecoded URIs.
-  # we'll see the decoded version as well
+  # skip undecoded URIs if the encoded bits shouldn't be.
+  # we'll see the decoded version as well.  see url_encode()
+  return if $uri =~ /\%(?:2[1-9a-fA-F]|[3-6][0-9a-fA-f]|7[0-9a-eA-E])/;
 
   # keep IPs intact
   if ($uri !~ /^\d+\.\d+\.\d+\.\d+$/) { 
@@ -822,13 +937,14 @@ sub uri_to_domain {
 }
 
 sub uri_list_canonify {
-  my(@uris) = @_;
+  my($redirector_patterns, @uris) = @_;
 
   # make sure we catch bad encoding tricks
   my @nuris = ();
   for my $uri (@uris) {
-    # we're interested in http:// and so on, skip mailto:
-    next if $uri =~ /^mailto:/i;
+    # we're interested in http:// and so on, skip mailto: and
+    # email addresses with no protocol
+    next if $uri =~ /^mailto:/i || $uri =~ /^[^:]*\@/;
 
     # sometimes we catch URLs on multiple lines
     $uri =~ s/\n//g;
@@ -904,6 +1020,20 @@ sub uri_list_canonify {
         push(@uris, $1);
       }
 
+      # resort to redirector pattern matching if the generic https? check
+      # doesn't result in a match -- bug 4176
+      else {
+	foreach (@{$redirector_patterns}) {
+	  if ("$proto$host$rest" =~ $_) {
+	    next unless defined $1;
+	    dbg("uri: parsed uri pattern: $_");
+	    dbg("uri: parsed uri found: $1 in redirector: $proto$host$rest");
+	    push (@uris, $1);
+	    last;
+	  }
+	}
+      }
+
       ########################
       ## TVD: known issue, if host has multiple combinations of the following,
       ## all permutations will be put onto @nuris.  shouldn't be an issue.
@@ -1034,7 +1164,7 @@ sub setuid_to_euid {
   my $touid = $>;
 
   if ($< != $touid) {
-    dbg ("changing real uid from $< to match effective uid $touid");
+    dbg("util: changing real uid from $< to match effective uid $touid");
     $< = $touid; # try the simple method first
 
     # bug 3586: Some perl versions, typically those on a BSD-based
@@ -1042,7 +1172,7 @@ sub setuid_to_euid {
     # can be changed.  So this is a kluge for us to get around the
     # typical spamd-ish behavior of: $< = 0, $> = someuid ...
     if ( $< != $touid ) {
-      dbg("initial attempt to change real uid failed, trying BSD workaround");
+      dbg("util: initial attempt to change real uid failed, trying BSD workaround");
 
       $> = $<;			# revert euid to ruid
       $< = $touid;		# change ruid to target
@@ -1051,7 +1181,8 @@ sub setuid_to_euid {
 
     # Check that we have now accomplished the setuid
     if ($< != $touid) {
-      die "setuid $< to $touid failed!";
+      # keep this fatal: it's a serious security problem if it fails
+      die "util: setuid $< to $touid failed!";
     }
   }
 }
@@ -1075,22 +1206,39 @@ sub helper_app_pipe_open_windows {
   return open ($fh, $cmd.'|');
 }
 
+sub force_die {
+  my ($msg) = @_;
+
+  # note use of eval { } scope in logging -- paranoia to ensure that a broken
+  # $SIG{__WARN__} implementation will not interfere with the flow of control
+  # here, where we *have* to die.
+  eval { warn $msg; };
+
+  POSIX::_exit(1);  # avoid END and destructor processing 
+  kill('KILL',$$);  # still kicking? die! 
+}
+
 sub helper_app_pipe_open_unix {
   my ($fh, $stdinfile, $duperr2out, @cmdline) = @_;
 
   # do a fork-open, so we can setuid() back
   my $pid = open ($fh, '-|');
   if (!defined $pid) {
-    die "cannot fork: $!";
+    # acceptable to die() here, calling code catches it
+    die "util: cannot fork: $!";
   }
 
   if ($pid != 0) {
     return $pid;          # parent process; return the child pid
   }
 
-  # else, child process.  go setuid...
+  # else, child process.  
+  # from now on, we cannot die(), as a parent-process eval { } scope
+  # could intercept it! use force_die() instead  (bug 4370, cmt 2)
+
+  # go setuid...
   setuid_to_euid();
-  dbg ("setuid: helper proc $$: ruid=$< euid=$>");
+  dbg("util: setuid: ruid=$< euid=$>");
 
   # now set up the fds.  due to some wierdness, we may have to ensure that we
   # *really* close the correct fd number, since some other code may have
@@ -1110,14 +1258,15 @@ sub helper_app_pipe_open_unix {
   if ($f != 0) {
     POSIX::close(0);
   }
-  open STDIN, "<$stdinfile" or die "cannot open $stdinfile: $!";
+
+  open (STDIN, "<$stdinfile") or force_die "util: cannot open $stdinfile: $!";
 
   # this should be impossible; if we just closed fd 0, UNIX
   # fd behaviour dictates that the next fd opened (the new STDIN)
   # will be the lowest unused fd number, which should be 0.
   # so die with a useful error if this somehow isn't the case.
   if (fileno(STDIN) != 0) {
-    die "setuid: oops: fileno(STDIN) [".fileno(STDIN)."] != 0";
+    force_die "util: setuid: oops: fileno(STDIN) [".fileno(STDIN)."] != 0";
   }
 
   # ensure STDOUT is open.  since we just created a pipe to ensure this, it has
@@ -1138,20 +1287,51 @@ sub helper_app_pipe_open_unix {
     if ($f != 2) {
       POSIX::close(2);
     }
-    open STDERR, ">&STDOUT" or die "dup STDOUT failed: $!";
+
+    open (STDERR, ">&STDOUT") or force_die "util: dup STDOUT failed: $!";
 
     # STDERR must be fd 2 to be useful to subprocesses! (bug 3649)
     if (fileno(STDERR) != 2) {
-      die "setuid: oops: fileno(STDERR) [".fileno(STDERR)."] != 2";
+      force_die "util: oops: fileno(STDERR) [".fileno(STDERR)."] != 2";
     }
   }
 
   exec @cmdline;
-  die "exec failed: $!";
+  warn "util: exec failed: $!";
+
+  # bug 4370: we really have to exit here; break any eval traps
+  POSIX::_exit(1);  # avoid END and destructor processing 
+  kill('KILL',$$);  # still kicking? die! 
+  die;  # must be a die() otherwise -w will complain
 }
 
 ###########################################################################
 
-sub dbg { Mail::SpamAssassin::dbg (@_); }
+# As "perldoc perlvar" notes, in perl 5.8.0, the concept of "safe" signal
+# handling was added, which means that signals cannot interrupt a running OP.
+# unfortunately, a regexp match is a single OP, so a psychotic m// can
+# effectively "hang" the interpreter as a result, and a $SIG{ALRM} handler
+# will never get called.
+#
+# However, by using "unsafe" signals, we can still interrupt that -- and
+# POSIX::sigaction can create an unsafe handler on 5.8.x.   So this function
+# provides a portable way to do that.
+
+sub trap_sigalrm_fully {
+  my ($handler) = @_;
+  if ($] < 5.008) {
+    # signals are always unsafe, just use %SIG
+    $SIG{ALRM} = $handler;
+  } else {
+    # may be using "safe" signals with %SIG; use POSIX to avoid it
+    POSIX::sigaction POSIX::SIGALRM(), new POSIX::SigAction $handler;
+  }
+}
+
+###########################################################################
 
 1;
+
+=back
+
+=cut
